"""Tools for automatic plotting with ``pyqtgraph``.

Contains a ``pyqtgraph``-specific implementation of :class:`.plot.base.FigureMaker`
(:class:`.FigureMaker`) and :class:`.plot.base.PlotWidget` (:class:`.AutoPlot`).
``FigureMaker`` can be used to quickly create a figure in a mostly automatic fashion
from known data.
For embedding in GUIs (such as plottr applications), ``AutoPlot`` is the main
object for plotting data automatically using ``pyqtgraph``.
"""

import logging
from pathlib import Path
import time
from dataclasses import dataclass
from typing import List, Optional, Any

import numpy as np
from pyqtgraph import mkPen

from plottr import QtWidgets, QtCore, Signal, Slot, \
    config_entry as getcfg
from plottr.data.datadict import DataDictBase
from .plots import Plot, PlotWithColorbar, PlotBase
from ..base import AutoFigureMaker as BaseFM, PlotDataType, \
    PlotItem, ComplexRepresentation, determinePlotDataType, \
    PlotWidgetContainer, PlotWidget

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
TIMESTRFORMAT = "%Y-%m-%dT%H%M%S"


class FigureWidget(QtWidgets.QWidget):
    """Widget that contains all plots generated by :class:`.FigureMaker`.

    Widget has a vertical layout, and plots can be added in a single column.
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        """Constructor for :class:`.FigureMakerWidget`.

        :param parent: parent widget.
        """
        super().__init__(parent=parent)

        self.subPlots: List[PlotBase] = []

        self.title = QtWidgets.QLabel(parent=self)
        self.title.setAlignment(QtCore.Qt.AlignHCenter)

        self.split = QtWidgets.QSplitter(parent=self)
        self.split.setOrientation(QtCore.Qt.Vertical)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        layout.addWidget(self.title)
        layout.addWidget(self.split)
        self.setLayout(layout)

        self.setTitle('')

    def addPlot(self, plot: PlotBase) -> None:
        """Add a :class:`.PlotBase` widget.

        :param plot: plot widget
        :param title: title of the plot
        """
        self.split.addWidget(plot)
        self.subPlots.append(plot)

    def clearAllPlots(self) -> None:
        """Clear all plot contents."""
        for p in self.subPlots:
            p.clearPlot()

    def deleteAllPlots(self) -> None:
        """Delete all subplot widgets."""
        for p in self.subPlots:
            p.deleteLater()
        self.subPlots = []

    def setTitle(self, title: str = '') -> None:
        self.title.setText(title)
        if len(title.strip()) == 0:
            self.title.setVisible(False)
        else:
            self.title.setVisible(True)


class FigureMaker(BaseFM):
    """``pyqtgraph`` implementation for :class:`.AutoFigureMaker`.

    Please see the base class for more general information.
    """

    # TODO: make scrollable when many figures (set min size)?
    # TODO: check for valid plot data

    def __init__(self, widget: Optional[FigureWidget] = None,
                 clearWidget: bool = True,
                 parentWidget: Optional[QtWidgets.QWidget] = None):
        """Constructor for :class:`.FigureMaker`.

        :param widget: a widget produced by an earlier use of this class
            may be supplied. This widget will then be used for generating the
            plot and avoids deleting and re-creating the widget if no structural
            changes to the plots is required (important case is live-plotting of
            data that keeps updating).
        :param clearWidget:
            if ``True`` force re-creation of all plot elements.
        :param parentWidget:
            parent of the main widget of FigureMaker.
        """
        super().__init__()

        self.widget: FigureWidget
        self.clearWidget = clearWidget

        if widget is None:
            self.widget = FigureWidget(parent=parentWidget)
        else:
            self.widget = widget

    # re-implementing to get correct type annotation.
    def __enter__(self) -> "FigureMaker":
        return self

    def subPlotFromId(self, subPlotId: int) -> PlotBase:
        """Get SubPlot from ID."""
        subPlots = self.subPlots[subPlotId].axes
        assert isinstance(subPlots, list) and len(subPlots) > 0 and \
               isinstance(subPlots[0], PlotBase)
        return subPlots[0]

    def makeSubPlots(self, nSubPlots: int) -> List[PlotBase]:
        """Create empty subplots in the widgets.

        If ``clearWidget`` was not set to ``True`` in the constructor,
        existing sub plot widgets are cleared, but not deleted and re-created.
        """
        plot: PlotBase
        if self.clearWidget:
            self.widget.deleteAllPlots()
            for i in range(nSubPlots):
                if max(self.dataDimensionsInSubPlot(i).values()) == 1:
                    plot = Plot(self.widget)
                    self.widget.addPlot(plot)
                elif max(self.dataDimensionsInSubPlot(i).values()) == 2:
                    plot = PlotWithColorbar(self.widget)
                    self.widget.addPlot(plot)
        else:
            self.widget.clearAllPlots()

        return self.widget.subPlots

    def formatSubPlot(self, subPlotId: int) -> None:
        """Set labels/legends for the given subplot."""
        if len(self.plotIdsInSubPlot(subPlotId)) == 0:
            return

        labels = self.subPlotLabels(subPlotId)
        subPlot = self.subPlotFromId(subPlotId)

        # label the x axis if there's only one x label
        if isinstance(subPlot, Plot):
            if len(set(labels[0])) == 1:
                subPlot.plot.setLabel("bottom", labels[0][0])

        if isinstance(subPlot, PlotWithColorbar):
            if len(set(labels[0])) == 1:
                subPlot.plot.setLabel("bottom", labels[0][0])

            if len(set(labels[1])) == 1:
                subPlot.plot.setLabel('left', labels[1][0])

            if len(set(labels[2])) == 1:
                subPlot.colorbar.setLabel('left', labels[2][0])

    def plot(self, plotItem: PlotItem) -> None:
        """Plot the given item."""
        if plotItem.plotDataType is PlotDataType.unknown:
            if len(plotItem.data) == 2:
                plotItem.plotDataType = PlotDataType.scatter1d
            elif len(plotItem.data) == 3:
                plotItem.plotDataType = PlotDataType.scatter2d

        if plotItem.plotDataType in [PlotDataType.scatter1d, PlotDataType.line1d]:
            self._1dPlot(plotItem)
        elif plotItem.plotDataType == PlotDataType.grid2d:
            self._colorPlot(plotItem)
        elif plotItem.plotDataType == PlotDataType.scatter2d:
            self._scatterPlot2d(plotItem)
        else:
            raise NotImplementedError('Cannot plot this data.')

    def _1dPlot(self, plotItem: PlotItem) -> None:
        colors = getcfg('main', 'pyqtgraph', 'line_colors', default=['r', 'b', 'g'])
        symbols = getcfg('main', 'pyqtgraph', 'line_symbols', default=['o'])
        symbolSize = getcfg('main', 'pyqtgraph', 'line_symbol_size', default=5)

        subPlot = self.subPlotFromId(plotItem.subPlot)

        assert len(plotItem.data) == 2
        x, y = plotItem.data

        color = colors[self.findPlotIndexInSubPlot(plotItem.id) % len(colors)]
        symbol = symbols[self.findPlotIndexInSubPlot(plotItem.id) % len(symbols)]

        if plotItem.plotDataType == PlotDataType.line1d:
            name = plotItem.labels[-1] if isinstance(plotItem.labels, list) else ''
            return subPlot.plot.plot(x.flatten(), y.flatten(), name=name,
                                     pen=mkPen(color, width=1), symbol=symbol, symbolBrush=color,
                                     symbolPen=None, symbolSize=symbolSize)
        else:
            name = plotItem.labels[-1] if isinstance(plotItem.labels, list) else ''
            return subPlot.plot.plot(x.flatten(), y.flatten(), name=name,
                                     pen=None, symbol=symbol, symbolBrush=color,
                                     symbolPen=None, symbolSize=symbolSize)

    def _colorPlot(self, plotItem: PlotItem) -> None:
        subPlot = self.subPlotFromId(plotItem.subPlot)
        assert isinstance(subPlot, PlotWithColorbar) and len(plotItem.data) == 3
        subPlot.setImage(*plotItem.data)

    def _scatterPlot2d(self, plotItem: PlotItem) -> None:
        subPlot = self.subPlotFromId(plotItem.subPlot)
        assert isinstance(subPlot, PlotWithColorbar) and len(plotItem.data) == 3
        subPlot.setScatter2d(*plotItem.data)


class AutoPlot(PlotWidget):
    """Widget for automatic plotting with pyqtgraph.

    Uses :class:`.FigureMaker` to produce subplots.
    """

    def __init__(self, parent: Optional[PlotWidgetContainer]) -> None:
        """Constructor for the pyqtgraph auto plot widget.

        :param parent: plot widget container
        """
        super().__init__(parent=parent)

        self.fmWidget: Optional[FigureWidget] = None
        self.figConfig: Optional[FigureConfigToolBar] = None
        self.figOptions: FigureOptions = FigureOptions()

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.setLayout(layout)
        self.setMinimumSize(*getcfg('main', 'pyqtgraph', 'minimum_plot_size',
                                    default=(400, 400)))

    def setData(self, data: Optional[DataDictBase]) -> None:
        """Uses :class:`.FigureMaker` to populate the plot(s).

        This method aims to re-use existing plotwidgets if possible -- i.e.,
        when the type of data (generally speaking) has not changed.
        If changes to the data structure are detected, then all subplot
        widgets are re-created from scratch, however.

        :param data: input data
        :return: ``None``
        """
        super().setData(data)
        if self.data is None:
            return

        fmKwargs = {}  # {'widget': self.fmWidget}
        dc = self.dataChanges
        if not dc['dataTypeChanged'] and not dc['dataStructureChanged']:
            fmKwargs['clearWidget'] = False
        else:
            fmKwargs['clearWidget'] = True
        self._plotData(**fmKwargs)

    def _plotData(self, **kwargs: Any) -> None:
        if self.data is None:
            return

        with FigureMaker(parentWidget=self, widget=self.fmWidget,
                         **kwargs) as fm:

            fm.complexRepresentation = self.figOptions.complexRepresentation
            fm.combineTraces = self.figOptions.combineLinePlots

            for dep in self.data.dependents():
                inds = self.data.axes(dep)
                dvals = self.data.data_vals(dep)
                pdt = determinePlotDataType(self.data.extract([dep]))
                plotId = fm.addData(
                    *[np.asanyarray(self.data.data_vals(n)) for n in inds] + [dvals],
                    labels=[str(self.data.label(n)) for n in inds] + [str(self.data.label(dep))],
                    plotDataType=pdt,
                )

        if self.fmWidget is None:
            self.fmWidget = fm.widget
            self.layout().addWidget(self.fmWidget)
            self.figConfig = FigureConfigToolBar(self.figOptions,
                                                 parent=self)
            self.layout().addWidget(self.figConfig)
            self.figConfig.optionsChanged.connect(self._refreshPlot)
            self.figConfig.figCopied.connect(self.onfigCopied)
            self.figConfig.figSaved.connect(self.onfigSaved)

        if self.data.has_meta('title'):
            self.fmWidget.setTitle(self.data.meta_val('title'))

    @Slot()
    def _refreshPlot(self) -> None:
        self._plotData()

    @Slot()
    def onfigCopied(self) -> None:
        """
        Gets triggered when figCopied signal is emitted from self.figConfig
        Copy the current figuremaker widget to the clipboard.
        """
        assert isinstance(self.fmWidget, FigureWidget)
        screenshot = self.fmWidget.grab(rectangle=QtCore.QRect(QtCore.QPoint(0, 0), QtCore.QSize(-1, -1)))
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setImage(screenshot.toImage())

    @Slot()
    def onfigSaved(self) -> None:
        """
        Gets triggered when figSaved signal is emitted from self.figConfig
        Save the current figuremaker widget to the data directory as a png
        with a timestamp at the front.
        """
        assert isinstance(self.fmWidget, FigureWidget)
        assert isinstance(self.data, DataDictBase)
        screenshot = self.fmWidget.grab(rectangle=QtCore.QRect(QtCore.QPoint(0, 0), QtCore.QSize(-1, -1)))
        path = Path(self.data.meta_val('title'))
        # add a timestamp here
        t = time.localtime()
        time_str = time.strftime(TIMESTRFORMAT, t)
        filename = time_str+'_'+str(path.stem)+'.png'
        screenshot.save(str(path.parent)+'/'+filename, format='PNG')

    # TODO: Allow for the option to choose filetypes and the name/directory


@dataclass
class FigureOptions:
    """Dataclass that describes the configuration options for the figure."""

    #: whether to plot all 1D traces into a single panel
    combineLinePlots: bool = False

    #: how to represent complex data
    complexRepresentation: ComplexRepresentation = ComplexRepresentation.realAndImag


class FigureConfigToolBar(QtWidgets.QToolBar):
    """Simple toolbar to configure the figure."""

    # TODO: find better config system that generates GUI automatically and
    #   links updates easier.

    #: Signal() -- emitted when options have been changed in the GUI.
    optionsChanged = Signal()
    #: Signal() -- emitted when the copy figure button has been pressed
    figCopied = Signal()
    #: Signal() -- emitted when the save figure button has been pressed
    figSaved = Signal()

    def __init__(self, options: FigureOptions,
                 parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Constructor.

        :param options: options object. GUI interaction will make changes
            in-place to this object.
        :param parent: parent Widget
        """
        super().__init__(parent)

        self.options = options

        combineLinePlots = self.addAction("Combine 1D")
        combineLinePlots.setCheckable(True)
        combineLinePlots.setChecked(self.options.combineLinePlots)
        combineLinePlots.triggered.connect(
            lambda: self._setOption('combineLinePlots',
                                    combineLinePlots.isChecked())
        )

        complexOptions = QtWidgets.QMenu(parent=self)
        complexGroup = QtWidgets.QActionGroup(complexOptions)
        complexGroup.setExclusive(True)
        for k in ComplexRepresentation:
            a = QtWidgets.QAction(k.label, complexOptions)
            a.setCheckable(True)
            complexGroup.addAction(a)
            complexOptions.addAction(a)
            a.setChecked(k == self.options.complexRepresentation)
        complexGroup.triggered.connect(
            lambda _a: self._setOption('complexRepresentation',
                                       ComplexRepresentation.fromLabel(_a.text()))
        )
        complexButton = QtWidgets.QToolButton()
        complexButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        complexButton.setText('Complex')
        complexButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        complexButton.setMenu(complexOptions)
        self.addWidget(complexButton)

        # Adding functionality to copy and save the graph
        self.copyFig = self.addAction('Copy Figure', self._copyFig)
        self.saveFig = self.addAction('Save Figure', self._saveFig)

    def _setOption(self, option: str, value: Any) -> None:
        setattr(self.options, option, value)
        self.optionsChanged.emit()

    def _copyFig(self) -> None:
        self.figCopied.emit()

    def _saveFig(self) -> None:
        self.figSaved.emit()



from typing import Tuple, Any, Optional, Union, Dict, List

import numpy as np
import lmfit

from plottr.analyzer.fitters.fitter_base import Fit, FitResult


class Cosine(Fit):
    @staticmethod
    def model(coordinates: np.ndarray, # type: ignore[override]
              A: float, f: float, phi: float, of: float) -> np.ndarray:
        """$A \cos(2 \pi f x + \phi) + of$"""
        return A * np.cos(2 * np.pi * coordinates * f + phi) + of

    @staticmethod
    def guess(coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
                 data: np.ndarray) -> Dict[str, float]:
        of = np.mean(data)
        A = (np.max(data) - np.min(data)) / 2.

        # Making sure that coordinates is ndarray.
        # Changing the type in the signature will create a different mypy error.
        assert isinstance(coordinates, np.ndarray)
        fft_val = np.fft.rfft(data)[1:]
        fft_frq = np.fft.rfftfreq(data.size,
                                  np.mean(coordinates[1:] - coordinates[:-1]))[1:]
        idx = np.argmax(np.abs(fft_val))
        f = fft_frq[idx]
        phi = np.angle(fft_val[idx])

        return dict(A=A, of=of, f=f, phi=phi)


class Exponential(Fit):
    @staticmethod
    def model(coordinates: np.ndarray, a: float, b: float) -> np.ndarray: # type: ignore[override]
        """ a * b ** x"""
        return a * b ** coordinates

    @staticmethod
    def guess(coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
                 data: np.ndarray) -> Dict[str, float]:
        return dict(a=1, b=2)



from typing import Any, Union, Tuple
import numpy as np

from ..base import Analysis, AnalysisResult


class FindMax(Analysis):
    """A simple example class to illustrate the concept."""

    def analyze(self, coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
                data: np.ndarray, *args: Any, **kwargs: Any) -> AnalysisResult:
        i = np.argmax(data)

        return AnalysisResult(
            dict(
                max_val=data[i],
                max_pos=coordinates[i]
            )
        )



def _get_version() -> str:
    from pathlib import Path

    import versioningit

    import plottr

    path = Path(plottr.__file__).parent
    return versioningit.get_version(project_dir=path.parent)


__version__ = "0.10.0"



from typing import Tuple, Any, Union, Dict

import numpy as np
import lmfit

from ..base import Analysis, AnalysisResult


class FitResult(AnalysisResult):

    def __init__(self, lmfit_result: lmfit.model.ModelResult):
        self.lmfit_result = lmfit_result
        self.params = lmfit_result.params

    def eval(self, *args: Any, **kwargs: Any) -> np.ndarray:
        return self.lmfit_result.eval(*args, **kwargs)


class Fit(Analysis):

    @staticmethod
    def model(*arg: Any, **kwarg: Any) -> np.ndarray:
        raise NotImplementedError

    def analyze(self, coordinates: Union[Tuple[np.ndarray, ...], np.ndarray], data: np.ndarray,
                dry: bool = False, params: Dict[str, Any] = {}, *args: Any, **fit_kwargs: Any) -> FitResult:
        model = lmfit.model.Model(self.model)

        _params = lmfit.Parameters()
        for pn, pv in self.guess(coordinates, data).items():
            _params.add(pn, value=pv)
        for pn, pv in params.items():
            if isinstance(pv, lmfit.Parameter):
                _params[pn] = pv
            else:
                _params[pn].set(value=pv)

        if dry:
            for pn, pv in _params.items():
                pv.set(vary=False)
        lmfit_result = model.fit(data, params=_params,
                                 coordinates=coordinates, **fit_kwargs)

        return FitResult(lmfit_result)

    @staticmethod
    def guess(coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
              data: np.ndarray) -> Dict[str, Any]:
        raise NotImplementedError



from typing import Tuple, Any, Optional, Union, Dict, List

import numpy as np
import lmfit
from plottr.analyzer.fitters.fitter_base  import Fit, FitResult


class T1_Decay(Fit):
    @staticmethod
    def model(coordinates: np.ndarray, amp: float, tau: float) -> np.ndarray: # type: ignore[override]
        """ amp * exp(-1.0 * x / tau)"""
        return amp * np.exp(-1.0 * coordinates / tau)
    @staticmethod
    def guess(coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
              data: np.ndarray) -> Dict[str, Any]:
        return dict(amp=1, tau=2)


class T2_Ramsey(Fit):
    @staticmethod
    def model(coordinates: np.ndarray, amp: float, tau: float, freq: float, phase: float) -> np.ndarray: # type: ignore[override]
        """ amp * exp(-1.0 * x / tau) * sin(2 * PI * freq * x + phase) """
        return amp * np.exp(-1.0 * coordinates / tau) * \
               np.sin(2 * np.pi * freq * coordinates + phase)

    @staticmethod
    def guess(coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
              data: np.ndarray) -> Dict[str, Any]:
        return dict(amp=1, tau=2, freq=3, phase=4)



from typing import Tuple, Any, Optional, Union, Dict, List
from collections import OrderedDict
from dataclasses import dataclass

import numpy as np
from matplotlib import pyplot as plt
import lmfit


class Parameter:

    def __init__(self, name: str, value: Any = None, **kw: Any):
        self.name = name
        self.value = value
        self._attrs = {}
        for k, v in kw.items():
            self._attrs[k] = v

    def __getattr__(self, key: str) -> Any:
        return self._attrs[key]


class Parameters(OrderedDict):
    """A collection of parameters"""

    def add(self, name: str, **kw: Any) -> None:
        """Add/overwrite a parameter in the collection."""
        self[name] = Parameter(name, **kw)


class AnalysisResult(object):

    def __init__(self, parameters: Dict[str, Union[Dict[str, Any], Any]]):
        self.params = Parameters()
        for k, v in parameters.items():
            if isinstance(v, dict):
                self.params.add(k, **v)
            else:
                self.params.add(k, value=v)

    def eval(self, *args: Any, **kwargs: Any) -> np.ndarray:
        """Analysis types that produce data (like filters or fits) should implement this.
        """
        raise NotImplementedError


class Analysis(object):
    """Basic analysis object.

    Parameters
    ----------
    coordinates
        may be a single 1d numpy array (for a single coordinate) or a tuple
        of 1d arrays (for multiple coordinates).
    data
        a 1d array of data
    """

    def __init__(self, coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
                 data: np.ndarray):
        """Constructor of `Analysis`. """
        self.coordinates = coordinates
        self.data = data

    def analyze(self, coordinates: Union[Tuple[np.ndarray, ...], np.ndarray], data: np.ndarray, *args: Any,
                **kwargs: Any) -> AnalysisResult:
        """Needs to be implemented by each inheriting class."""
        raise NotImplementedError

    def run(self, *args: Any, **kwargs: Any) -> AnalysisResult:
        return self.analyze(self.coordinates, self.data, **kwargs)


# def analyze(analysis_class: Analysis, coordinates: Union[Tuple[np.ndarray, ...], np.ndarray],
#             data: np.ndarray, **kwarg: Any) -> AnalysisResult:
#     analysis = analysis_class(coordinates, data)
#     return analysis.run(**kwarg)




import sys
import os
import pkgutil
import importlib
from importlib import reload, import_module
import warnings
from typing import Dict, Optional, Type, Callable, Tuple, Any, List, Union
import inspect
from dataclasses import dataclass
import numbers
from types import ModuleType

import lmfit
from lmfit import Parameter as lmParameter, Parameters as lmParameters

from plottr import QtGui, QtCore, Slot, Signal, QtWidgets
from plottr.analyzer import fitters
from plottr.analyzer.fitters.fitter_base import Fit, FitResult

from ..data.datadict import DataDictBase
from .node import Node, NodeWidget, updateOption, updateGuiFromNode

__author__ = 'Chao Zhou'
__license__ = 'MIT'


def reload_module_get_model(module: ModuleType) -> Tuple[ModuleType, Dict[str, type]]:
    '''Gather the model classes in the the fitting module file
    :return : a dictionary that contains all the model classed in the module
    '''
    def is_Fit_subclass(cls: Type[Fit]) -> bool:
        """ check if a class is the subclass of analyzer.fitters.fitter_base.Fit
        """
        try:
            if issubclass(cls, Fit) and (cls is not Fit):
                return True
            else:
                return False
        except TypeError:
            return False
    # reload the module (this will clear the class cache)
    try:
        del sys.modules[module.__name__]
    except:
        pass
    module = import_module(module.__name__)

    model_classes = inspect.getmembers(module, is_Fit_subclass)
    model_dict = {}
    for mc in model_classes:
        model_dict[mc[0]] = mc[1]
    return module, model_dict

def get_modules_in_pkg(pkg: ModuleType) -> Dict[str, ModuleType]:
    '''Gather the fitting modules in a package
    '''
    modules = {}
    for importer, modname, ispkg in pkgutil.iter_modules(pkg.__path__):
        if modname != "fitter_base":
            module_ = import_module('.'+modname, pkg.__name__)
            try:
                del sys.modules[module_.__name__]
            except:
                pass
            module_ = import_module('.'+modname, pkg.__name__)
            modules[modname] = module_
    return modules


INITIAL_MODULES = get_modules_in_pkg(fitters)
#TODO: this requires putting the modules in the init of fitters, there should
# be a better way to do this.

# OPEN_MODULE_ICON = QtGui.QIcon(QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.SP_DirOpenIcon))
# REFRESH_MODULE_ICON = QtGui.QIcon(QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.SP_BrowserReload))

MAX_FLOAT = sys.float_info.max

DEBUG = 1



@dataclass
class FittingOptions:
    model: Type[Fit]
    parameters: lmParameters
    dry_run: bool = False


class FittingGui(NodeWidget):
    """ Gui for controlling the fitting function and the initial guess of
    fitting parameters.
    """
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, node: Optional[Node] = None):

        super().__init__(parent)
        self.input_options: Optional[FittingOptions] = None # fitting option in dataIn
        self.live_update = False
        self.dry_run = False
        self.param_signals: List[QtCore.pyqtBoundSignal] = []
        self.fitting_modules = INITIAL_MODULES
        self.my_layout = QtWidgets.QGridLayout()
        self.setLayout(self.my_layout)

        # fitting module widgets
        module_sel_widget = QtWidgets.QWidget()
        module_sel_grid = QtWidgets.QGridLayout()
        # model function selection widget
        self.module_combo = self.addModuleComboBox()
        self.module_combo.currentTextChanged.connect(self.moduleUpdate)
        self.module_combo.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                        QtWidgets.QSizePolicy.Fixed)
        module_sel_grid.addWidget(self.module_combo, 0, 0)
        # refresh module button
        # refresh_button = QtWidgets.QPushButton(REFRESH_MODULE_ICON, "")
        refresh_button = QtWidgets.QPushButton("↻")
        refresh_button.clicked.connect(self.moduleRefreshClicked)
        refresh_button.setSizePolicy(QtWidgets.QSizePolicy.Fixed,
                                     QtWidgets.QSizePolicy.Fixed)
        module_sel_grid.addWidget(refresh_button, 0, 1)
        # add module button
        # open_button = QtWidgets.QPushButton(OPEN_MODULE_ICON,"")
        open_button = QtWidgets.QPushButton("+")
        open_button.clicked.connect(self.add_user_module)
        open_button.setSizePolicy(QtWidgets.QSizePolicy.Fixed,
                                  QtWidgets.QSizePolicy.Fixed)
        module_sel_grid.addWidget(open_button, 0, 2)
        module_sel_widget.setLayout(module_sel_grid)
        self.my_layout.addWidget(module_sel_widget, 0, 0)


        # model list widget
        self.model_list = QtWidgets.QListWidget()
        self.moduleUpdate(self.module_combo.currentText())
        self.model_list.currentItemChanged.connect(self.modelChanged)

        # function description window
        self.model_doc_box = QtWidgets.QLineEdit("")
        self.model_doc_box.setReadOnly(True)

        # splitter1
        splitter1 = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        splitter1.addWidget(self.model_list)
        splitter1.addWidget(self.model_doc_box)

        # parameter table
        self.param_table = QtWidgets.QTableWidget(0, 4)
        self.param_table.setHorizontalHeaderLabels([
            'fix', 'initial guess', 'lower bound', 'upper bound'])
        self.param_table.horizontalHeader(). \
            setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        self.param_table.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                        QtWidgets.QSizePolicy.Expanding)

        # splitter2
        splitter2 = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        splitter2.addWidget(splitter1)
        splitter2.addWidget(self.param_table)
        self.my_layout.addWidget(splitter2, 1, 0)

        # fitting update options
        self.update_option_widget = self.addUpdateOptions()
        self.my_layout.addWidget(self.update_option_widget, 2, 0)


        # getter and setter
        self.optGetters['fitting_options'] = self.fittingOptionGetter
        self.optSetters['fitting_options'] = self.fittingOptionSetter

    def addModuleComboBox(self) -> QtWidgets.QComboBox:
        """ Set up the model function drop down manual widget.
        """
        combo = QtWidgets.QComboBox()
        combo.setEditable(False)
        for module_name in self.fitting_modules:
            combo.addItem(module_name)
        return combo

    @Slot(str)
    def moduleUpdate(self, current_module_name: str) -> None:
        if DEBUG:
            print ("GUI...: ", "moduleUpdate called. Updating",current_module_name)
        self.model_list.clear()
        current_module = self.fitting_modules[current_module_name]
        # update the module in console and get models in new module
        new_module, new_models = reload_module_get_model(current_module)
        for model_name in new_models:
            self.model_list.addItem(model_name)
        self.fitting_modules[current_module_name] = new_module
        # debug-------------------------------------------
        '''
        # fitters.generic_functions.Cosine.pp(1)
        try:
            test = self.fitting_modules[current_module_name].Exponential2
            print("Exponential2 is here!!!!!")
        except :
            print("No Exponential2 :( ")
        '''
        #-------------------------------------------------

    @Slot()
    def moduleRefreshClicked(self) -> None:
        self.moduleUpdate(self.module_combo.currentText())

    def add_user_module(self) -> None:
        mod_file = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Open file',fitters.__path__[0], "Python files (""*.py)")[0]
        if (mod_file is None) or mod_file[-3:] !=".py":
            return
        mod_name = mod_file.split('/')[-1][:-3]
        mod_dir = '\\'.join(mod_file.split('/')[:-1])
        # load the selected module
        sys.path.append(mod_dir)
        user_module = import_module(mod_name, mod_dir)

        # check if there is already a module with the same name
        if mod_name in self.fitting_modules:
            for existing_module in self.fitting_modules.values():
                if user_module.__file__ == existing_module.__file__:
                    # a module that already exists is loaded
                    print("a module that already exists is loaded")
                    self.module_combo.setCurrentText(mod_name)
                    self.moduleUpdate(mod_name)
                    return
            # a different module whose name is the same as one of the
            # existing modules is loaded
            print("a different module whose name is the same as one of "
                  "the existing modules is loaded")
            mod_name += f"({mod_dir})"

        self.fitting_modules[mod_name] = user_module
        self.module_combo.addItem(mod_name)
        self.module_combo.setCurrentText(mod_name)

    @Slot(QtWidgets.QListWidgetItem, QtWidgets.QListWidgetItem)
    def modelChanged(self,
                     current: QtWidgets.QListWidgetItem,
                     previous: QtWidgets.QListWidgetItem) -> None:
        """ Process a change in fit model selection.
        Will update the parameter table based on the new selection.
        """
        if current is None:
            print ("No model selected")
            self.model_doc_box.setText("")
            self.param_table.setRowCount(0)
            return
        current_module = self.fitting_modules[self.module_combo.currentText()]
        model_cls = getattr(current_module, current.text())
        self.updateParamTable(model_cls)
        self.model_doc_box.setText(model_cls.model.__doc__)

    def updateParamTable(self, model_cls: Type[Fit]) -> None:
        """ Update the parameter table based on the current model selection.
        :param model_cls: the current selected fitting model class
        """
        # flush param table
        self.param_table.setRowCount(0)
        # rebuild param table based on the selected model function
        func = model_cls.model
        # assume the first variable is the independent variable
        params = list(inspect.signature(func).parameters)[1:]
        self.param_table.setRowCount(len(params))
        self.param_table.setVerticalHeaderLabels(params)
        # generate fix, initial guess, lower/upper bound option GUIs for each
        # parameter
        self.param_signals = []
        for idx, name in enumerate(params):
            fixParamCheck = self._paramFixCheck()
            fixParamCheck.setStyleSheet("margin-left:15%; margin-right:5%;")
            initialGuessBox = OptionSpinbox(1.0, self)
            lowerBoundBox = NumberInput(None, self)
            upperBoundBox = NumberInput(None, self)
            lowerBoundBox.newTextEntered.connect(initialGuessBox.setMinimum)
            upperBoundBox.newTextEntered.connect(initialGuessBox.setMaximum)

            # gather the param change signals for enabling live update
            self.param_signals.extend([fixParamCheck.stateChanged,
                                       initialGuessBox.valueChanged,
                                       lowerBoundBox.newTextEntered,
                                       upperBoundBox.newTextEntered])
            # put param options into table
            self.param_table.setCellWidget(idx, 0, fixParamCheck)
            self.param_table.setCellWidget(idx, 1, initialGuessBox)
            self.param_table.setCellWidget(idx, 2, lowerBoundBox)
            self.param_table.setCellWidget(idx, 3, upperBoundBox)

        self.changeParamLiveUpdate(self.live_update)

    def _paramFixCheck(self, default_value: bool = False) -> QtWidgets.QCheckBox:
        """generate a push checkbox for the parameter fix option.
        :param default_value : param is fixed by default or not
        :returns: a checkbox widget
        """
        widget = QtWidgets.QCheckBox('')
        widget.setChecked(default_value)
        widget.setToolTip("when fixed, the parameter will be fixed to the "
                          "initial guess value during fitting")
        return widget

    def addUpdateOptions(self) -> QtWidgets.QWidget:
        ''' Add check box & buttons that control the fitting update policy.
        '''
        update_option_widget = QtWidgets.QWidget()
        grid = QtWidgets.QGridLayout()
        # when checked, fitting will update after each change of fitting model
        # or parameter option
        liveUpdateCheck = QtWidgets.QCheckBox('Live Update')
        grid.addWidget(liveUpdateCheck, 0, 0)
        # update fitting on-demand
        updateFitButton = QtWidgets.QPushButton("Fit")
        grid.addWidget(updateFitButton, 0, 1)
        # guess fitting parameters and show the dry run result
        guessParamButton = QtWidgets.QPushButton("Guess Parameters")
        grid.addWidget(guessParamButton, 0, 2)
        # reload the fitting options that come from the data
        reloadInputOptButton = QtWidgets.QPushButton("Reload Input Option")
        grid.addWidget(reloadInputOptButton, 0, 3)

        @Slot(QtCore.Qt.CheckState)
        def setLiveUpdate(live: QtCore.Qt.CheckState) -> None:
            ''' connect/disconnects the changing signal of each fitting
            option to signalAllOptions slot
            '''
            if live == QtCore.Qt.Checked:
                self.model_list.currentItemChanged.connect(
                    self._signalAllOptions)
                self.changeParamLiveUpdate(True)
                self.live_update = True
            else:
                try:
                    self.model_list.currentItemChanged.disconnect(
                        self._signalAllOptions)
                except TypeError:
                    pass
                self.changeParamLiveUpdate(False)
                self.live_update = False

        @Slot()
        def reloadInputOption() -> None:
            if DEBUG:
                print("GUI...: ", "reload input option")
            self.fittingOptionSetter(self.input_options)

        @Slot()
        def setGuessParam() -> None:
            if DEBUG:
                print("GUI...: ", "setGuessParam called, setting fitting parameter to guess")
            if self.model_list.currentItem() is None:
                return
            self.dry_run = True
            self.signalAllOptions()

        liveUpdateCheck.stateChanged.connect(setLiveUpdate)
        updateFitButton.pressed.connect(self.signalAllOptions)
        guessParamButton.pressed.connect(setGuessParam)
        reloadInputOptButton.pressed.connect(reloadInputOption)
        reloadInputOptButton.setToolTip('reload the fitting options stored '
                                        'in the input data')

        update_option_widget.setLayout(grid)
        return update_option_widget

    def changeParamLiveUpdate(self, enable: bool) -> None:
        ''' connect/disconnects the changing signal of each fitting param
        option to signalAllOptions slot
        :param enable: connect/disconnect when enable is True/False.
        '''
        if enable:
            for psig in self.param_signals:
                psig.connect(self._signalAllOptions)
        else:
            for psig in self.param_signals:
                try:
                    psig.disconnect(self._signalAllOptions)
                except TypeError:
                    pass

    def getCurrentModel(self) -> Optional[Type[Fit]]:
        """
        get the current model selected
        """
        current_module = self.fitting_modules[self.module_combo.currentText()]
        model_selected = self.model_list.currentItem()
        if model_selected is None:
            return None
        model = getattr(current_module, model_selected.text())
        return model


    def fittingOptionGetter(self) -> Optional[FittingOptions]:
        """ get all the fitting options and put them into a dictionary
        """
        if DEBUG:
            print("GUI...: ", 'getter in gui called')
        # get the current model selected
        model = self.getCurrentModel()
        if model is None:
            return None
        # get the parameters for current model
        parameters = lmParameters()
        for i in range(self.param_table.rowCount()):
            table_item = self.param_table.verticalHeaderItem(i)
            assert isinstance(table_item, QtWidgets.QTableWidgetItem)
            param_name = table_item.text()
            param = lmParameter(param_name)
            item0 = self.param_table.cellWidget(i, 0)
            item1 = self.param_table.cellWidget(i, 1)
            item2 = self.param_table.cellWidget(i, 2)
            item3 = self.param_table.cellWidget(i, 3)
            assert isinstance(item0, QtWidgets.QCheckBox)
            assert isinstance(item1, OptionSpinbox)
            assert isinstance(item2, NumberInput)
            assert isinstance(item3, NumberInput)
            param.vary = not item0.isChecked()
            param.value = item1.value()
            param.min = item2.value()
            param.max = item3.value()
            parameters[param_name] = param

        fitting_options = FittingOptions(model, parameters, self.dry_run)
        if DEBUG:
            print("GUI...: ", 'getter in gui got', fitting_options)
        return fitting_options

    def fittingOptionSetter(self, fitting_options: Optional[FittingOptions]) -> None:
        """ Set all the fitting options
        """
        if DEBUG:
            print("GUI...: ", 'setter in gui called')
        if fitting_options is None:
            return
        # set the model in gui
        model = fitting_options.model
        if DEBUG:
            print("GUI...: ", f"setter trying to set model to {model}")
        # try to find the module that contains the model first
        module_exist = False
        for mdu_name, mdu in self.fitting_modules.items():
            if mdu.__file__ == inspect.getsourcefile(model):
                self.module_combo.setCurrentText(mdu_name)
                if DEBUG:
                    print("GUI...: ", f"setter set module to {mdu_name}")
                module_exist = True
                break
        # set the model in model list
        if module_exist:
            model_cls_name = model.__qualname__
            find_mdls = self.model_list.findItems(model_cls_name,
                                                  QtCore.Qt.MatchExactly)
            if len(find_mdls) == 1:
                self.model_list.setCurrentItem(find_mdls[0])
            else:
                if DEBUG:
                    print("GUI...: ", "unexpected Error when trying to find the module")
                    print(model_cls_name)
                else:
                    raise NameError("GUI...: ", "unexpected Error when trying to find the module")
        else:
            #TODO: fix this, add the new model to gui
            raise NotImplementedError("auto add new model to gui")

        # set the parameter table in gui
        for i in range(self.param_table.rowCount()):
            table_item = self.param_table.verticalHeaderItem(i)
            assert isinstance(table_item, QtWidgets.QTableWidgetItem)
            param_name = table_item.text()
            param_options = fitting_options.parameters[param_name]

            item0 = self.param_table.cellWidget(i, 0)
            item1 = self.param_table.cellWidget(i, 1)
            item2 = self.param_table.cellWidget(i, 2)
            item3 = self.param_table.cellWidget(i, 3)
            assert isinstance(item0, QtWidgets.QCheckBox)
            assert isinstance(item1, OptionSpinbox)
            assert isinstance(item2, NumberInput)
            assert isinstance(item3, NumberInput)

            item0.setChecked(not param_options.vary)
            item1.setValue(param_options.value)
            item2.setValue(param_options.min)
            item3.setValue(param_options.max)
        self.dry_run = fitting_options.dry_run

    def _signalAllOptions(self, *args: Any) -> None:
        # to make the signalAllOptions accept signals w/ multi args
        if DEBUG:
            print("GUI...: ", "signal all option change")
        if self.model_list.currentItem() is not None:
            self.signalAllOptions()

    @updateGuiFromNode
    def setDefaultFit(self, fitting_options: FittingOptions) -> None:
        ''' set the gui to the fitting options in the input data
        '''
        if DEBUG:
            print("GUI...: ", f'updateGuiFromNode setDefault got {fitting_options}')
        if self.fittingOptionGetter() is None:
            self.fittingOptionSetter(fitting_options)
        self.input_options = fitting_options

    @updateGuiFromNode
    def setGuessParam(self, fitting_options: FittingOptions) -> None:
        """ set the parameter to the guess parameter from guess function
        """
        if DEBUG:
            print("GUI...: ", f'updateGuiFromNode setGuessParam got {fitting_options}')
        self.fittingOptionSetter(fitting_options)


class OptionSpinbox(QtWidgets.QDoubleSpinBox):
    """A spinBox widget for parameter options
    :param default_value : default value of the option
    """

    # TODO: Support easier input for large numbers
    def __init__(self, default_value: float = 1.0, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        self.setRange(-1 * MAX_FLOAT, MAX_FLOAT)
        self.setValue(default_value)

    def setMaximum(self, maximum: str) -> None:  # type: ignore[override]
        try:
            value = eval(maximum)
        except:
            value = MAX_FLOAT
        if isinstance(value, float):
            super().setMaximum(value)
        else:
            super().setMaximum(MAX_FLOAT)

    def setMinimum(self, minimum: str) -> None:  # type: ignore[override]
        try:
            value = eval(minimum)
        except:
            value = -1 * MAX_FLOAT
        if isinstance(value, float):
            super().setMinimum(value)
        else:
            super().setMinimum(-1 * MAX_FLOAT)


class NumberInput(QtWidgets.QLineEdit):
    """A text edit widget that checks whether its input can be read as a
    number.
    This is copied form the parameter GUI that Wolfgang wrote for the
    parameter manager gui.
    """
    newTextEntered = Signal(str)

    def __init__(self, default_value: Union[numbers.Number, None], parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        self.setValue(default_value)
        self.editingFinished.connect(self.emitNewText)

    def value(self) -> Optional[numbers.Number]:
        try:
            value = eval(self.text())
        except:
            return None
        if isinstance(value, numbers.Number):
            return value
        else:
            return None

    def setValue(self, value: Union[numbers.Number, None]) -> None:
        self.setText(str(value))

    def emitNewText(self) -> None:
        self.newTextEntered.emit(self.text())


# ================= Node ==============================
class FittingNode(Node):
    uiClass = FittingGui
    nodeName = "Fitter"
    default_fitting_options = Signal(object)
    guess_fitting_options = Signal(object)

    def __init__(self, name: str):
        super().__init__(name)
        self._fitting_options: Optional[FittingOptions] = None

    def process(self, dataIn: Optional[DataDictBase] = None) -> Optional[Dict[str, Optional[DataDictBase]]]:
        return self.fitting_process(dataIn)

    @property
    def fitting_options(self) -> Optional[FittingOptions]:
        return self._fitting_options

    @fitting_options.setter  # type: ignore[misc] # https://github.com/python/mypy/issues/1362
    @updateOption('fitting_options')
    def fitting_options(self, opt: Optional[FittingOptions]) -> None:
        if isinstance(opt, FittingOptions) or opt is None:
            self._fitting_options = opt
        else:
            raise TypeError('Wrong fitting options')

    def fitting_process(self, dataIn: Optional[DataDictBase] = None) -> Optional[Dict[str, Optional[DataDictBase]]]:
        if dataIn is None:
            return None

        if len(dataIn.axes()) > 1 or len(dataIn.dependents()) > 1:
            return dict(dataOut=dataIn)

        dataIn_opt = dataIn.get('__fitting_options__')
        dataOut = dataIn.copy()

        # no fitting option selected in gui
        if self.fitting_options is None:
            if dataIn_opt is not None:
                self._fitting_options = dataIn_opt
            else:
                return dict(dataOut=dataOut)

        if dataIn_opt is not None:
            if DEBUG:
                print("NODE>>>: ", "Emit initial option from node!", dataIn_opt)
            self.default_fitting_options.emit(dataIn_opt)

        # fitting
        if DEBUG:
            print("NODE>>>: ", f"node got fitting option {self.fitting_options}")

        axname = dataIn.axes()[0]
        x = dataIn.data_vals(axname)
        y = dataIn.data_vals(dataIn.dependents()[0])

        assert isinstance(self.fitting_options, FittingOptions)
        fit = self.fitting_options.model(x, y)
        if self.fitting_options.dry_run:
            guess_params = lmParameters()
            for pn, pv in fit.guess(x, y).items():
                guess_params.add(pn, value=pv)
            guess_opts = FittingOptions(self.fitting_options.model,
                                        guess_params, False)
            self.guess_fitting_options.emit(guess_opts)
            if DEBUG:
                print("NODE>>>: ", f"guess param in node. Emit guess_opts: {guess_opts}")
            # show dry run result
            fit_result = fit.run(dry=True)
            result_y = fit_result.eval(coordinates=x)
            dataOut['guess'] = dict(values=result_y, axes=[axname, ])
        else:
            fit_result = fit.run(params=self.fitting_options.parameters)
            assert isinstance(fit_result, FitResult)
            lm_result = fit_result.lmfit_result
            if lm_result.success:
                dataOut['fit'] = dict(values=lm_result.best_fit, axes=[axname,])
                dataOut.add_meta('info', lm_result.fit_report())

        return dict(dataOut=dataOut)

    def setupUi(self) -> None:
        super().setupUi()
        assert isinstance(self.ui, FittingGui)
        self.default_fitting_options.connect(self.ui.setDefaultFit)
        self.guess_fitting_options.connect(self.ui.setGuessParam)

